# $Id: thumb.py,v 1.2 2004/05/08 18:39:30 stephen Exp $


"""Template for a thumbnail generation program.  This provides a Python
class which you extend to generate a thumbnail image for a type of file.

An example is included which generates thumbnails for text files.  A more
useful implemetation is VideoThumbnail.
"""

import os, sys
import md5

import rox

# Class for thumbnail programs
class Thumbnailler:
    """Base class for programs which generate thumbnails.  You should
    override the method get_image() to create the image.  You can also
    override post_process_image()."""
    def __init__(self, name, fname, use_wdir=False, debug=False):
        """Initialise the thumbnailler.
        name - name of the program
        fname - a string to use in generated temp file names
        use_wdir - if true then use a temp directory to store files
        debug - if false then suppress most error messages
        """
        self.name=name
        self.fname=fname
        self.use_wdir=use_wdir
        self.debug=debug

    def run(self, inname, outname=None, rsize=96):
        """Generate the thumbnail from the file
        inname - source file
        outname - path to store thumbnail image
        rsize - maximum size of thumbnail (in either axis)
        """
        if not outname:
            uri='file://'+inname
            tmp=md5.new(uri).digest()
            leaf=''
            for c in tmp:
                leaf+='%02x' % ord(c)
            outname=os.path.join(os.environ['HOME'], '.thumbnails',
                                 'normal', leaf+'.png')
        elif not os.path.isabs(outname):
            outname=os.path.abspath(outname)

        if self.use_wdir:
            self.make_working_dir()

        try:
            img=self.get_image(inname, rsize)
            ow=img.get_width()
            oh=img.get_height()        
            img=self.process_image(img, rsize)
            self.store_image(img, inname, outname, ow, oh)
        except:
            self.report_exception()

        if self.use_wdir:
            self.remove_working_dir()

    def get_image(self, inname, rsize):
        """Method you must define for your thumbnailler to do anything"""
        raise "Thumbnail not implemented"

    def process_image(self, img, rsize):
        """Take the raw image and scale it to the correct size.
        Returns the result of scaling img and passing it to
        post_process_image()"""
        ow=img.get_width()
        oh=img.get_height()
        if ow>oh:
            s=float(rsize)/float(ow)
        else:
            s=float(rsize)/float(oh)
        w=int(s*ow)
        h=int(s*oh)

        img=img.scale_simple(w, h, rox.g.gdk.INTERP_BILINEAR)

        return self.post_process_image(img, w, h)

    def post_process_image(self, img, w, h):
        """Perform some post-processing on the image.
        img - gdk-pixbuf of the image
        w - width
        h - height
        Return: modified image
        The default implementation just returns the image unchanged."""
        return img

    def store_image(self, img, inname, outname, ow, oh):
        """Store the thumbnail image it the correct location, adding
        the extra data required by the thumbnail spec."""
        s=os.stat(inname)

        img.save(outname+self.fname, 'png',
             {'tEXt::Thumb::Image::Width': str(ow),
              'tEXt::Thumb::Image::Height': str(oh),
              "tEXt::Thumb::Size": str(s.st_size),
              "tEXt::Thumb::MTime": str(s.st_mtime),
              'tEXt::Thumb::URI': 'file://'+inname,
              'tEXt::Software': self.name})
        os.rename(outname+self.fname, outname)
        
    def make_working_dir(self):
        """Create the temporary directory and change into it."""
        self.work_dir=os.path.join('/tmp',
                                       '%s.%d' % (self.fname, os.getpid()))
        #print work_dir
        try:
            os.makedirs(self.work_dir)
        except:
            self.report_exception()
            self.work_dir=None
            return

        self.old_dir=os.getcwd()
        os.chdir(self.work_dir)
        
    def remove_working_dir(self):
        """Remove our temporary directory, after changing back to the
        previous one"""
        if not self.work_dir:
            return
        
        os.chdir(self.old_dir)

        for f in os.listdir(self.work_dir):
            path=os.path.join(self.work_dir, f)

            try:
                os.remove(path)
            except:
                self.report_exception()

        try:
            os.rmdir(self.work_dir)
        except:
            self.report_exception()
        
    def report_exception(self):
        """Report an exception (if debug enabled)"""
        if self.debug<1:
            return
        #exc=sys.exc_info()[:2]
        #sys.stderr.write('%s: %s %s\n' % (sys.argv[0], exc[0], exc[1]))
        rox.report_exception()
        



# Most of the common code needed by ROX applications is in ROX-Lib2.
# Except this code, which is needed to find ROX-Lib2 in the first place!

# Just make sure you run findrox.version() before importing anything inside
# ROX-Lib2...

import os, sys
from os.path import exists
import string

def version(major, minor, micro):
	"""Find ROX-Lib2, with a version >= (major, minor, micro), and
	add it to sys.path. If version is missing or too old, either
	prompt the user, or (if possible) upgrade it automatically.
	If 'rox' is already in PYTHONPATH, just use that (assume the injector
	is being used)."""
	try:
		import rox
	except ImportError:
		pass
	else:
		#print "Using ROX-Lib in PYTHONPATH"
		if (major, minor, micro) > rox.roxlib_version:
			print >>sys.stderr, "WARNING: ROX-Lib version " \
				"%d.%d.%d requested, but using version " \
				"%d.%d.%d from %s" % \
				(major, minor, micro,
				 rox.roxlib_version[0],
				 rox.roxlib_version[1],
				 rox.roxlib_version[2],
				 rox.__file__)
		return

	if not os.getenv('ROXLIB_DISABLE_ZEROINSTALL') and os.path.exists('/uri/0install/rox.sourceforge.net'):
		# We're using ZeroInstall. Good :-)
		zpath = '/uri/0install/rox.sourceforge.net/lib/ROX-Lib2/' \
			'latest-2'
		if not os.path.exists(zpath):
			os.system('0refresh rox.sourceforge.net')
			assert os.path.exists(zpath)
		vs = os.readlink(zpath).split('-')[-1]
		v = tuple(map(int, vs.split('.')))
		if v < (major, minor, micro):
			if os.system('0refresh rox.sourceforge.net'):
				report_error('Using ROX-Lib in Zero Install, but cached version (%s) is too old (need %d.%d.%d) and updating failed (is zero-install running?)' % (vs, major, minor, micro))
		sys.path.append(zpath + '/python')
		return
	try:
		path = os.environ['LIBDIRPATH']
		paths = string.split(path, ':')
	except KeyError:
		paths = [os.environ['HOME'] + '/lib',
			 '/usr/local/lib', '/usr/lib' ]

	for p in paths:
		p = os.path.join(p, 'ROX-Lib2')
		if exists(p):
			# TODO: check version is new enough
			sys.path.append(os.path.join(p, 'python'))
			import rox
			if major == 1 and minor == 9 and micro < 10:
				return	# Can't check version
			if not hasattr(rox, 'roxlib_version'):
				break
			if (major, minor, micro) <= rox.roxlib_version:
				return	# OK
	report_error("This program needs ROX-Lib2 (version %d.%d.%d) " % \
		(major, minor, micro) + "to run.\n" + \
		"I tried all of these places:\n\n" + \
		string.join(paths, '\n') + '\n\n' + \
		"ROX-Lib2 is available from:\n" + \
		"http://rox.sourceforge.net")

def report_error(err):
	"Write 'error' to stderr and, if possible, display a dialog box too."
	try:
		sys.stderr.write('*** ' + err + '\n')
	except:
		pass
	try:
		import pygtk; pygtk.require('2.0')
		import gtk; g = gtk
	except:
		import gtk
		win = gtk.GtkDialog()
		message = gtk.GtkLabel(err + 
				'\n\nAlso, pygtk2 needs to be present')
		win.set_title('Missing ROX-Lib2')
		win.set_position(gtk.WIN_POS_CENTER)
		message.set_padding(20, 20)
		win.vbox.pack_start(message)

		ok = gtk.GtkButton("OK")
		ok.set_flags(gtk.CAN_DEFAULT)
		win.action_area.pack_start(ok)
		ok.connect('clicked', gtk.mainquit)
		ok.grab_default()
		
		win.connect('destroy', gtk.mainquit)
		win.show_all()
		gtk.mainloop()
	else:
		box = g.MessageDialog(None, g.MESSAGE_ERROR, 0,
					g.BUTTONS_OK, err)
		box.set_title('Missing ROX-Lib2')
		box.set_position(g.WIN_POS_CENTER)
		box.set_default_response(g.RESPONSE_OK)
		box.run()
	sys.exit(1)



#!/usr/bin/env python
#
# $Id: vidthumb.py,v 1.14 2005/10/08 10:57:22 stephen Exp $

"""Generate thumbnails for video files.  This must be called as
      vidthumb.py source_file destination_thumbnail maximum_size
where maximum_size is the maximum width and height.

This uses mplayer to grab a single frame as a .png file, then gdk-pixbuf (via
pygtk 1.99.x) to generate a thumbnail of that frame.  A film strip effect is
added to mark the thumbnail as from a video file.

To use this in ROX-Filer, get a recent version then run VideoThumbnail and
click on "Install handlers".  Generating a video thumbnail takes a lot longer
than a simple image file, but ROX-Filer remains responsive while it is
being processed.
"""

import os, sys
import md5
import rox, rox.mime
import pango

import thumb

# Defaults
import options

outname=None
rsize=options.tsize.int_value
take_first=options.take_first.int_value

# Width of the film strip effect to put at each side
bwidth=options.ssize.int_value

# How best to select the frame for a given type.  Worked out by trial and error
first_by_types={
    'video/mpeg': True,
    'video/x-ms-wmv': False
}

def binaryInPath(b):
    for path in os.environ["PATH"].split(":"):
        if os.access(os.path.join(path, b), os.R_OK | os.X_OK):
            return True
    return False

def execute_return_err(cmd):
    errmsg=""
    cin, cout=os.popen4(cmd, 'r')
    cin.close()
    for l in cout:
        errmsg+=l
    cout.close()
    #cerr.close()
    if debug: print cmd, errmsg
    if errmsg:
        return errmsg
        
debug=os.environ.get('VIDTHUMB_DEBUG', 0)

class VidThumbNail(thumb.Thumbnailler):
    """Generate thumbnail for video files understood by totem"""
    def __init__(self, debug=False):
        """Initialize Video thumbnailler"""
        thumb.Thumbnailler.__init__(self, 'VideoThumbnail', 'vidthumb',
                                    True, debug)

    def failed_image(self, rsize, tstr):
        if debug: print 'failed_image', self, rsize, tstr
        if not tstr:
            tstr='Error!'
        w=rsize
        h=rsize/4*3
        try:
            p=rox.g.gdk.Pixbuf(rox.g.gdk.COLORSPACE_RGB, False, 8, w, h)
        except:
            sys.exit(2)
        if debug: print p

        pixmap, mask=p.render_pixmap_and_mask()
        cmap=pixmap.get_colormap()
        gc=pixmap.new_gc(foreground=cmap.alloc_color('black'))
        gc.set_foreground(gc.background)

        if debug: print gc, gc.foreground
        pixmap.draw_rectangle(gc, rox.g.TRUE, 0, 0, w, h)
        
        gc.set_foreground(cmap.alloc_color('red'))
        dummy=rox.g.Window()
        layout=dummy.create_pango_layout(tstr)
        if w>40:
            layout.set_width((w-10)*pango.SCALE)
            #layout.set_wrap(pango.WRAP_CHAR)
        pixmap.draw_layout(gc, 10, 4, layout)
        if debug: print pixmap

        self.add_time=False
        
        return p.get_from_drawable(pixmap, cmap, 0, 0, 0, 0, -1, -1)

    def check_executable(cls):
        try:
            return binaryInPath(cls._binary)
        except Exception, e:
            if debug:
                print e
            return False
    check_executable = classmethod(check_executable)
        
        

class VidThumbTotem(VidThumbNail):
    """Generate thumbnail for video files understood by totem"""
    _binary = "totem-video-thumbnailer"
    def __init__(self, debug=False):
        """Initialize Video thumbnailler"""
        VidThumbNail.__init__(self, debug)


    def get_image(self, inname, rsize):
        outfile = os.path.join(self.work_dir, "out.png")
        if options.scale.int_value:
            cmd = 'totem-video-thumbnailer -s %i \"%s\" %s' % (rsize, inname, outfile)
        else:
            cmd = 'totem-video-thumbnailer \"%s\" %s' % (inname, outfile)
        errmsg=execute_return_err(cmd)
        if not os.path.exists(outfile):
            return self.failed_image(rsize, errmsg)

        # Now we load the raw image in
        return rox.g.gdk.pixbuf_new_from_file(outfile)


class VidThumbMPlayer(VidThumbNail):
    """Generate thumbnail for video files understood by mplayer"""
    _binary = "mplayer"
    def __init__(self, debug=False):
        """Initialize Video thumbnailler"""
        VidThumbNail.__init__(self, debug)

        self.add_time=options.time_label.int_value

    def post_process_image(self, img, w, h):
        """Add the optional film strip effect"""

        if not options.sprocket.int_value and not self.add_time:
            return img
        
        pixmap, mask=img.render_pixmap_and_mask()
        cmap=pixmap.get_colormap()
        gtk=rox.g
        gc=pixmap.new_gc(foreground=cmap.alloc_color('black'))
            
        if options.sprocket.int_value:
            # Draw the film strip effect
            pixmap.draw_rectangle(gc, True, 0, 0, 8, h)
            pixmap.draw_rectangle(gc, True, w-8, 0, 8, h)

            gc.set_foreground(cmap.alloc_color('#DDD'))
            for y in range(1, h, 8):
                pixmap.draw_rectangle(gc, True, 2, y, 4, 4)
                pixmap.draw_rectangle(gc, True, w-8+2, y, 4, 4)

        if self.add_time and self.total_time:
            secs=self.total_time
            hours=int(secs/3600)
            if hours>0:
                secs-=hours*3600
            mins=int(secs/60)
            if mins>0:
                secs-=mins*60
            tstr='%d:%02d:%02d' % (hours, mins, secs)
            if debug: print tstr
            dummy=gtk.Window()
            layout=dummy.create_pango_layout(tstr)
            
            gc.set_foreground(cmap.alloc_color('black'))
            pixmap.draw_layout(gc, 11, 5, layout)
            gc.set_foreground(cmap.alloc_color('white'))
            pixmap.draw_layout(gc, 10, 4, layout)
            if debug: print layout
                               
        return img.get_from_drawable(pixmap, cmap, 0, 0, 0, 0, -1, -1)


    def get_image(self, inname, rsize):
        """Generate the raw image from the file.  We run mplayer (twice)
        to do the hard work."""
        def get_length(fname):
            """Get the length in seconds of the source. """
            # -frames 0 might be needed on debian systems
            unused, inf, junk=os.popen3(
                'mplayer -frames 0 -vo null -ao null -identify "%s"' % fname,
                'r')

            for l in inf.readlines():
                # print l[:10]
                if l[:10]=='ID_LENGTH=':
                    return float(l.strip()[10:])

        def write_frame(fname, pos):
            """Return filename of a single frame from the source, taken 
            from pos seconds into the video"""

            # Ask for 2 frames.  Seems to work better
            cmd='mplayer -really-quiet -vo png -ss %f -frames 2 -nosound -noloop "%s"' % (pos, fname)
            cmd+=' > /dev/null 2>&1'

            # If we have 2 frames ignore the first and return the second, else
            # if we have 1 return it.  Otherwise mplayer couldn't cope and we
            # return None
            def frame_ok(ofile):
                if debug: print 'look for', ofile
                try:
                    os.stat(ofile)
                except:
                    if debug: print 'exception', sys.exc_info()[:2]
                    if debug: os.system('pwd')
                    if debug: os.system('ls -al')
                    return False
                return True

            if debug: print cmd
            os.system(cmd)

            mtype=rox.mime.get_type(fname)
            #print >>sys.stderr, mtype, first_by_types
            try:
                #print >>sys.stderr, str(mtype)
                first=first_by_types[str(mtype)]
            except:
                #print >>sys.stderr, 'oops'
                first=take_first
            #print >>sys.stderr, first
            
            if first:
                id=1
            else:
                id=2
            ofile='%08d.png' % id
            if debug: print ofile
            if not frame_ok(ofile):
                if not first:
                    id=1
                    ofile='%08d.png' % id
                    if not frame_ok(ofile):
                        ofile=None
                else:
                    ofile=None
                
            return ofile

        try:
            vlen=get_length(inname)
        except:
            self.report_exception()
            return self.failed_image(rsize, 'Bad length')

        self.total_time=vlen
        if debug: print vlen
    
        # Select a frame 5% of the way in, but not more than 60s  (Long files
        # usually have a fade in).
        pos=vlen*0.05
        if pos>60:
            pos=60

        frfname=write_frame(inname, pos)
        if debug: print inname, pos, frfname
        if frfname is None:
            return self.failed_image(rsize, 'Bad or missing frame file')

        # Now we load the raw image in

        return rox.g.gdk.pixbuf_new_from_file(frfname)

thumbnailers = {"mplayer": VidThumbMPlayer,
                "totem" : VidThumbTotem }

# Process command line args.  Although the filer always passes three args,
# let the last two default to something sensible to allow use outside
# the filer.
def main(argv):
    """Process command line args.  Although the filer always passes three args,
    let the last two default to something sensible to allow use outside
    the filer."""
    global rsize
    
    #print argv
    inname=argv[0]
    try:
        outname=argv[1]
    except:
        pass
    try:
        rsize=int(argv[2])
    except:
        pass
    
    if not os.path.isabs(inname):
        inname=os.path.abspath(inname)

    # Out file name is based on MD5 hash of the URI
    if not outname:
        uri='file://'+inname
        tmp=md5.new(uri).digest()
        leaf=''
        for c in tmp:
            leaf+='%02x' % ord(c)
        outname=os.path.join(os.environ['HOME'], '.thumbnails', 'normal',
                         leaf+'.png')
    elif not os.path.isabs(outname):
        outname=os.path.abspath(outname)
    #print inname, outname, rsize
    thumbC = thumbnailers.pop(options.generator.value)
    if not thumbC.check_executable():
        origbin = thumbC._binary
        for (name, cls) in thumbnailers.iteritems():
            if cls.check_executable():
                thumbC = cls
                msg = """VideoThumbnail could not find the program "%s", but another thumbnail generator is available.
                
Should "%s" be used from now on?"""
                if rox.confirm(msg % (origbin, thumbC._binary), rox.g.STOCK_YES):
                    options.generator.value = name
                    rox.app_options.save()
                    pass
                break
        else:
            msg = """VideoThumbnail could not find any usable thumbnail generator.
            
You need to install either MPlayer (http://www.mplayerhq.hu)
or Totem (http://www.gnome.org/projects/totem/)."""
            rox.croak(msg)
            
    thumb = thumbC(options.report.int_value)
    thumb.run(inname, outname, rsize)

        
def configure():
    """Configure the app"""
    options.edit_options()
    rox.mainloop()

if __name__=='__main__':
    main(sys.argv[1:])



"""Handle the options for the video thumbnailler"""

# $Id: options.py,v 1.8 2005/10/08 10:57:22 stephen Exp $

import os

import rox
import rox.options
import rox.OptionsBox

rox.setup_app_options('VideoThumbnail', site='kerofin.demon.co.uk')


tsize=rox.options.Option('tsize', 128)
sprocket=rox.options.Option('sprocket', 1)
ssize=rox.options.Option('ssize', 8)
time_label=rox.options.Option('time', 0)
report=rox.options.Option('report', 0)
take_first=rox.options.Option('take_first', False)
generator=rox.options.Option('generator', 'mplayer')
scale=rox.options.Option('scale', False)

rox.app_options.notify()

def install_button_handler(*args):
    import rox.mime_handler
    try:
        rox.mime_handler.install_from_appinfo()
    except:
        rox.report_exception()

def build_install_button(box, node, label):
    #print box, node, label
    button = rox.g.Button(label)
    box.may_add_tip(button, node)
    button.connect('clicked', install_button_handler)
    return [button]
rox.OptionsBox.widget_registry['install-button'] = build_install_button

        
def edit_options():
    rox.edit_options()
